from django.contrib import admin

from .models import (
    Cohort,
    SignupCodeCohort,
    Survey,
    SurveyAnswer,
    SurveyInstance,
    SurveyQuestion,
    SurveyQuestionChoice,
    UserCohort,
    WaitingListEntry
)


class WaitingListEntryAdmin(admin.ModelAdmin):
    
    list_display = ["email", "created"]
    search_fields = ["email"]


class SignupCodeCohortInline(admin.TabularInline):
    
    model = SignupCodeCohort


class UserCohortInline(admin.TabularInline):
    
    model = UserCohort


class SurveyInstanceAdmin(admin.ModelAdmin):
    
    model = SurveyInstance
    list_display = ["survey", "email", "created"]
    
    def survey(self, obj):
        return obj.survey.label
    
    def email(self, obj):
        return obj.entry.email
    
    def created(self, obj):
        return obj.entry.created


class SurveyAnswerAdmin(admin.ModelAdmin):
    
    model = SurveyAnswer
    list_display = ["survey", "email", "question_label", "value", "value_boolean", "created"]
    
    def survey(self, obj):
        return obj.instance.survey.label
    
    def email(self, obj):
        return obj.instance.entry.email
    
    def question_label(self, obj):
        return obj.question.question


class SurveyQuestionChoiceInline(admin.TabularInline):
    
    model = SurveyQuestionChoice


class SurveyQuestionAdmin(admin.ModelAdmin):
    
    model = SurveyQuestion
    list_display = ["survey", "question", "kind", "required"]
    inlines = [SurveyQuestionChoiceInline]
    
    def survey(self, obj):
        return obj.survey.label


admin.site.register(WaitingListEntry, WaitingListEntryAdmin)
admin.site.register(
    Cohort,
    inlines=[
        SignupCodeCohortInline,
        UserCohortInline,
    ]
)

admin.site.register(
    Survey,
    list_display=["label", "active"]
)
admin.site.register(SurveyAnswer, SurveyAnswerAdmin)
admin.site.register(SurveyInstance, SurveyInstanceAdmin)
admin.site.register(SurveyQuestion, SurveyQuestionAdmin)



from django import template

from waitinglist.forms import WaitingListEntryForm


register = template.Library()


@register.assignment_tag
def waitinglist_entry_form():
    """
    Get a (new) form object to post a new comment.
    
    Syntax::
    
        {% waitinglist_entry_form as [varname] %}
    
    """
    return WaitingListEntryForm()



from django.test import TestCase

from .forms import SurveyForm
from .models import (
    Survey,
    SurveyQuestion,
    WaitingListEntry
)


class SurveyTests(TestCase):
    
    def setUp(self):
        self.survey = Survey.objects.create(
            label="My Test Survey"
        )
        self.entry = WaitingListEntry.objects.create(email="pinax@awesome.com")
        self.ice_cream_question = self.survey.questions.create(
            question="What is your favorite ice cream flavor?",
            kind=SurveyQuestion.TEXT_FIELD,
            help_text="(e.g. Vanilla, Strawberry, Chocolate)",
            required=True
        )
        self.summer_question = self.survey.questions.create(
            question="What did you do last summer?",
            kind=SurveyQuestion.TEXT_AREA,
            required=False
        )
        self.season_question = self.survey.questions.create(
            question="What is your favorite season?",
            kind=SurveyQuestion.RADIO_CHOICES,
            required=True
        )
        self.spring = self.season_question.choices.create(
            label="Spring"
        )
        self.summer = self.season_question.choices.create(
            label="Summer"
        )
        self.fall = self.season_question.choices.create(
            label="Fall"
        )
        self.winter = self.season_question.choices.create(
            label="Winter"
        )
        self.city_question = self.survey.questions.create(
            question="Select all the cities you have visited",
            kind=SurveyQuestion.CHECKBOX_FIELD,
            required=True
        )
        self.boston = self.city_question.choices.create(
            label="Boston"
        )
        self.denver = self.city_question.choices.create(
            label="Denver"
        )
        self.nashville = self.city_question.choices.create(
            label="Nashville"
        )
        self.danville = self.city_question.choices.create(
            label="Danville"
        )
        self.golf_question = self.survey.questions.create(
            question="Do you like golf?",
            kind=SurveyQuestion.BOOLEAN_FIELD,
            required=True
        )
    
    def test_create_second_survey(self):
        Survey.objects.create(label="Another test survey")
        self.assertEquals(Survey.objects.count(), 2)
        self.assertEquals(Survey.objects.filter(active=False).count(), 1)
        self.assertEquals(Survey.objects.filter(active=True).count(), 1)
    
    def test_survey_form_creation(self):
        form = SurveyForm(survey=self.survey)
        self.assertTrue(len(form.fields), 5)
    
    def test_survey_form_invalid(self):
        form = SurveyForm(
            data={
                self.ice_cream_question.name: "Strawberry"
            },
            survey=self.survey
        )
        self.assertFalse(form.is_valid())
    
    def test_survey_form_valid(self):
        form = SurveyForm(
            data={
                self.ice_cream_question.name: "Strawberry",
                self.summer_question.name: "Swam in the lake",
                self.season_question.name: self.summer.pk,
                self.city_question.name: [self.nashville.pk],
                self.golf_question.name: True
            },
            survey=self.survey
        )
        self.assertTrue(form.is_valid())
    
    def test_survey_form_save(self):
        form = SurveyForm(
            data={
                self.ice_cream_question.name: "Strawberry",
                self.summer_question.name: "Swam in the lake",
                self.season_question.name: self.summer.pk,
                self.city_question.name: [self.nashville.pk, self.boston.pk],
                self.golf_question.name: True
            },
            survey=self.survey
        )
        self.assertTrue(form.is_valid())
        form.save(self.entry.surveyinstance)
        answers = self.entry.surveyinstance.answers.all()
        self.assertEquals(answers.count(), 5)
        self.assertEquals(answers.get(question=self.ice_cream_question).value, "Strawberry")
        self.assertEquals(answers.get(question=self.summer_question).value, "Swam in the lake")
        self.assertEquals(answers.get(question=self.season_question).value, self.summer.label)
        self.assertTrue(
            self.nashville.label in answers.get(question=self.city_question).value
        )
        self.assertTrue(
            self.boston.label in answers.get(question=self.city_question).value
        )
        self.assertTrue(answers.get(question=self.golf_question).value_boolean)



from django import forms

from .models import Cohort, SurveyAnswer, SurveyQuestion, WaitingListEntry
from .signals import answered_survey


class WaitingListEntryForm(forms.ModelForm):
    
    class Meta:
        model = WaitingListEntry
    
    def clean_email(self):
        value = self.cleaned_data["email"]
        try:
            entry = WaitingListEntry.objects.get(email=value)
        except WaitingListEntry.DoesNotExist:
            return value
        else:
            raise forms.ValidationError(
                "The email address %(email)s already registered on %(date)s." % {
                    "email": value,
                    "date": entry.created.strftime("%m/%d/%y"),
                }
            )
    
    def __init__(self, *args, **kwargs):
        super(WaitingListEntryForm, self).__init__(*args, **kwargs)
        self.fields["email"].widget.attrs["placeholder"] = "your@email.com"
        self.fields["email"].label = ""


class CohortCreate(forms.ModelForm):
    
    class Meta:
        model = Cohort
        exclude = ["created"]


class SurveyForm(forms.Form):
    
    def __init__(self, *args, **kwargs):
        self.survey = kwargs.pop("survey")
        super(SurveyForm, self).__init__(*args, **kwargs)
        for question in self.survey.questions.all():
            self.fields[question.name] = question.form_field()
    
    def save(self, instance):
        for question in self.survey.questions.all():
            answer = SurveyAnswer.objects.create(instance=instance, question=question)
            value = self.cleaned_data[question.name]
            if question.kind == SurveyQuestion.RADIO_CHOICES:
                answer.value = value.label
            elif question.kind == SurveyQuestion.CHECKBOX_FIELD:
                answer.value = ", ".join([x.label for x in value])
            elif question.kind == SurveyQuestion.BOOLEAN_FIELD:
                answer.value_boolean = value
            else:
                answer.value = value
            answer.save()
        answered_survey.send(sender=self, instance=instance)



from django.conf.urls.defaults import patterns, url
from django.views.generic import TemplateView


urlpatterns = patterns("",
    url(r"^list_signup/$", "waitinglist.views.list_signup", name="waitinglist_list_signup"),
    url(r"^ajax_list_signup/$", "waitinglist.views.ajax_list_signup", name="waitinglist_ajax_list_signup"),
    url(r"^survey/thanks/$", TemplateView.as_view(template_name="waitinglist/thanks.html"), name="waitinglist_thanks"),
    url(r"^survey/(?P<code>.*)/$", "waitinglist.views.survey", name="waitinglist_survey"),
    url(r"^success/$", TemplateView.as_view(template_name="waitinglist/success.html"), name="waitinglist_success"),
    url(r"^cohorts/$", "waitinglist.views.cohort_list", name="waitinglist_cohort_list"),
    url(r"^cohorts/create/$", "waitinglist.views.cohort_create", name="waitinglist_cohort_create"),
    url(r"^cohorts/cohort/(\d+)/$", "waitinglist.views.cohort_detail", name="waitinglist_cohort_detail"),
    url(r"^cohorts/cohort/(\d+)/add_member/$", "waitinglist.views.cohort_member_add", name="waitinglist_cohort_member_add"),
    url(r"^cohorts/cohort/(\d+)/send_invitations/$", "waitinglist.views.cohort_send_invitations", name="waitinglist_cohort_send_invitations"),
)



import json

from django.core.urlresolvers import reverse
from django.http import Http404, HttpResponse
from django.shortcuts import render, redirect, get_object_or_404
from django.template import RequestContext
from django.template.loader import render_to_string
from django.views.decorators.http import require_POST

from django.contrib.auth.models import User

from account.models import SignupCode

from .forms import WaitingListEntryForm, CohortCreate, SurveyForm
from .models import WaitingListEntry, Cohort, SignupCodeCohort, SurveyInstance
from .signals import signed_up


@require_POST
def ajax_list_signup(request):
    form = WaitingListEntryForm(request.POST)
    if form.is_valid():
        entry = form.save()
        signed_up.send(sender=ajax_list_signup, entry=entry)
        try:
            data = {
                "location": reverse("waitinglist_survey", args=[entry.surveyinstance.code])
            }
        except SurveyInstance.DoesNotExist:
            data = {
                "html": render_to_string("waitinglist/_success.html", {
                },  context_instance=RequestContext(request))
            }
    else:
        data = {
            "html": render_to_string("waitinglist/_list_signup.html", {
                "form": form,
            },  context_instance=RequestContext(request))
        }
    return HttpResponse(json.dumps(data), mimetype="application/json")


def list_signup(request, post_save_redirect=None):
    if request.method == "POST":
        form = WaitingListEntryForm(request.POST)
        if form.is_valid():
            entry = form.save()
            signed_up.send(sender=list_signup, entry=entry)
            try:
                post_save_redirect = reverse("waitinglist_survey", args=[entry.surveyinstance.code])
            except SurveyInstance.DoesNotExist:
                pass
            if post_save_redirect is None:
                post_save_redirect = reverse("waitinglist_success")
            if not post_save_redirect.startswith("/"):
                post_save_redirect = reverse(post_save_redirect)
            return redirect(post_save_redirect)
    else:
        form = WaitingListEntryForm()
    ctx = {
        "form": form,
    }
    return render(request, "waitinglist/list_signup.html", ctx)


def survey(request, code):
    instance = get_object_or_404(SurveyInstance, code=code)
    if request.method == "POST":
        form = SurveyForm(request.POST, survey=instance.survey)
        if form.is_valid():
            form.save(instance)
            return redirect("waitinglist_thanks")
    else:
        form = SurveyForm(survey=instance.survey)
    return render(request, "waitinglist/survey.html", {"form": form})


def cohort_list(request):
    
    if not request.user.is_staff:
        raise Http404()
    
    ctx = {
        "cohorts": Cohort.objects.order_by("-created")
    }
    return render(request, "cohorts/cohort_list.html", ctx)


def cohort_create(request):
    
    if not request.user.is_staff:
        raise Http404()
    
    if request.method == "POST":
        form = CohortCreate(request.POST)
        
        if form.is_valid():
            cohort = form.save()
            return redirect("waitinglist_cohort_detail", cohort.id)
    else:
        form = CohortCreate()
    
    ctx = {
        "form": form,
    }
    return render(request, "cohorts/cohort_create.html", ctx)


def cohort_detail(request, pk):
    
    if not request.user.is_staff:
        raise Http404()
    
    cohort = get_object_or_404(Cohort, pk=pk)
    
    # people who are NOT invited or on the site already
    waiting_list = WaitingListEntry.objects.exclude(
        email__in=SignupCode.objects.values("email")
    ).exclude(
        email__in=User.objects.values("email")
    )
    
    ctx = {
        "cohort": cohort,
        "waiting_list": waiting_list,
    }
    return render(request, "cohorts/cohort_detail.html", ctx)


def cohort_member_add(request, pk):
    
    if not request.user.is_staff:
        raise Http404()
    
    cohort = Cohort.objects.get(pk=pk)
    
    if "invite_next" in request.POST:
        try:
            N = int(request.POST["invite_next"])
        except ValueError:
            return redirect("waitinglist_cohort_detail", cohort.id)
        # people who are NOT invited or on the site already
        waiting_list = WaitingListEntry.objects.exclude(
            email__in=SignupCode.objects.values("email")
        ).exclude(
            email__in=User.objects.values("email")
        )
        emails = waiting_list.values_list("email", flat=True)[:N]
    else:
        email = request.POST["email"].strip()
        if email:
            emails = [email]
        else:
            emails = []
    
    for email in emails:
        if not SignupCode.objects.filter(email=email).exists():
            signup_code = SignupCode.create(email=email, max_uses=1, expiry=730)
            signup_code.save()
            SignupCodeCohort.objects.create(signup_code=signup_code, cohort=cohort)
    
    return redirect("waitinglist_cohort_detail", cohort.id)


def cohort_send_invitations(request, pk):
    
    if not request.user.is_staff:
        raise Http404()
    
    cohort = Cohort.objects.get(pk=pk)
    cohort.send_invitations()
    
    return redirect("waitinglist_cohort_detail", cohort.id)



import collections
import hashlib

from django import forms
from django.conf import settings
from django.db import models
from django.db.models import Max
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.template.defaultfilters import slugify
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

from django.contrib.auth.models import User

from account.models import SignupCode, SignupCodeResult
from account.signals import user_signed_up


SURVEY_SECRET = getattr(settings, "WAITINGLIST_SURVEY_SECRET", settings.SECRET_KEY)


class WaitingListEntry(models.Model):
    
    email = models.EmailField(_("email address"), unique=True)
    created = models.DateTimeField(_("created"), default=timezone.now, editable=False)
    
    class Meta:
        verbose_name = _("waiting list entry")
        verbose_name_plural = _("waiting list entries")


@receiver(post_save, sender=WaitingListEntry)
def handle_waitinglistentry_save(sender, **kwargs):
    if kwargs.get("created"):
        try:
            survey = Survey.objects.get(active=True)
            SurveyInstance.objects.create(
                survey=survey,
                entry=kwargs.get("instance")
            )
        except Survey.DoesNotExist:
            pass


class Survey(models.Model):
    
    label = models.CharField(max_length=100, unique=True)
    active = models.BooleanField(default=True)
    
    def __unicode__(self):
        return self.label
    
    def save(self, *args, **kwargs):
        if self.active:
            Survey.objects.filter(active=True).update(active=False)
        return super(Survey, self).save(*args, **kwargs)


class SurveyInstance(models.Model):
    
    survey = models.ForeignKey(Survey, related_name="instances")
    entry = models.OneToOneField(WaitingListEntry)
    code = models.CharField(max_length=200, unique=True)
    
    def generate_hash(self):
        return hashlib.md5(self.entry.email + SURVEY_SECRET).hexdigest()
    
    def save(self, *args, **kwargs):
        self.code = self.generate_hash()
        return super(SurveyInstance, self).save(*args, **kwargs)


class SurveyQuestion(models.Model):
    
    TEXT_FIELD = 0
    TEXT_AREA = 1
    RADIO_CHOICES = 2
    CHECKBOX_FIELD = 3
    BOOLEAN_FIELD = 4
    
    FIELD_TYPE_CHOICES = [
        (TEXT_FIELD, "text field"),
        (TEXT_AREA, "textarea"),
        (RADIO_CHOICES, "radio choices"),
        (CHECKBOX_FIELD, "checkbox field (can select multiple answers"),
        (BOOLEAN_FIELD, "boolean field")
    ]
    
    survey = models.ForeignKey(Survey, related_name="questions")
    question = models.TextField()
    kind = models.IntegerField(choices=FIELD_TYPE_CHOICES)
    help_text = models.TextField(blank=True)
    ordinal = models.IntegerField(blank=True)
    required = models.BooleanField()
    
    class Meta:
        unique_together = [
            ("survey", "question")
        ]
        ordering = ["ordinal"]
    
    @property
    def name(self):
        return slugify(self.question)
    
    def form_field(self):
        kwargs = dict(
            label=self.question,
            help_text=self.help_text,
            required=self.required
        )
        field_class = forms.CharField
        
        if self.kind == SurveyQuestion.TEXT_AREA:
            kwargs.update({"widget": forms.Textarea()})
        elif self.kind == SurveyQuestion.RADIO_CHOICES:
            field_class = forms.ModelChoiceField
            kwargs.update({"widget": forms.RadioSelect(), "queryset": self.choices.all()})
        elif self.kind == SurveyQuestion.CHECKBOX_FIELD:
            field_class = forms.ModelMultipleChoiceField
            kwargs.update({"widget": forms.CheckboxSelectMultiple(), "queryset": self.choices.all()})
        elif self.kind == SurveyQuestion.BOOLEAN_FIELD:
            field_class = forms.BooleanField
        
        return field_class(**kwargs)
    
    def save(self, *args, **kwargs):
        if not self.pk:
            max_ordinal = self.survey.questions.aggregate(
                Max("ordinal")
            )["ordinal__max"] or 0
            self.ordinal = max_ordinal + 1
        return super(SurveyQuestion, self).save(*args, **kwargs)


class SurveyQuestionChoice(models.Model):
    question = models.ForeignKey(SurveyQuestion, related_name="choices")
    label = models.CharField(max_length=100)
    
    def __unicode__(self):
        return self.label


class SurveyAnswer(models.Model):
    
    instance = models.ForeignKey(SurveyInstance, related_name="answers")
    question = models.ForeignKey(SurveyQuestion, related_name="answers")
    value = models.TextField(blank=True)
    value_boolean = models.NullBooleanField(blank=True)
    created = models.DateTimeField(_("created"), default=timezone.now, editable=False)


Member = collections.namedtuple("Member", ["email", "signup_code", "user", "invited"])


class Cohort(models.Model):
    
    name = models.CharField(_("name"), max_length=35)
    created = models.DateTimeField(_("created"), default=timezone.now, editable=False)
    
    def members(self):
        members = []
        for scc in self.signupcodecohort_set.select_related():
            try:
                scr = SignupCodeResult.objects.get(signup_code=scc.signup_code_id)
            except SignupCodeResult.DoesNotExist:
                user = None
            else:
                user = scr.user
            members.append(
                Member(
                    scc.signup_code.email,
                    scc.signup_code,
                    user,
                    bool(scc.signup_code.sent)
                )
            )
        return members
    
    def member_counts(self):
        members = self.members()
        return {
            "total": len(members),
            "users": len([m for m in members if m.user is not None]),
            "pending": len([m.signup_code for m in members if not m.invited]),
        }
    
    def send_invitations(self):
        for sc in [m.signup_code for m in self.members() if not m.invited]:
            sc.send()
    
    def __unicode__(self):
        return self.name


class SignupCodeCohort(models.Model):
    """
    fetch cohort of a given signup code
        SignupCodeCohort.objects.select_related("cohort").get(signup_code__code="abc").cohort
        
    list of people waiting NOT on the site already or invited
        WaitingListEntry.objects.exclude(email__in=SignupCode.objects.values("email")).exclude(email__in=User.objects.values("email"))
    """
    signup_code = models.OneToOneField(SignupCode)
    cohort = models.ForeignKey(Cohort)


class UserCohort(models.Model):
    """
    Upon signup we create an instance of this model associating the new user and their cohort
    """
    user = models.OneToOneField(User)
    cohort = models.ForeignKey(Cohort)


@receiver(user_signed_up)
def handle_user_signup(sender, **kwargs):
    signup_code = kwargs["form"].cleaned_data["code"]
    # fetch the cohort for the signup code
    qs = SignupCodeCohort.objects.select_related("cohort")
    try:
        cohort = qs.get(signup_code__code=signup_code).cohort
        # create a UserCohort for user association to a cohort
        UserCohort.objects.create(user=kwargs["user"], cohort=cohort)
    except SignupCodeCohort.DoesNotExist:
        pass



from django.conf import settings
from django.core.mail import EmailMessage
from django.core.management.base import BaseCommand
from django.template.loader import render_to_string

from django.contrib.sites.models import Site

from ...models import WaitingListEntry, Survey


class Command(BaseCommand):
    
    help = "Email links to survey instances for those that never saw a survey"
    
    def handle(self, *args, **options):
        survey = Survey.objects.get(active=True)
        entries = WaitingListEntry.objects.filter(surveyinstance__isnull=True)
        
        for entry in entries:
            instance = survey.instances.create(entry=entry)
            site = Site.objects.get_current()
            protocol = getattr(settings, "DEFAULT_HTTP_PROTOCOL", "http")
            ctx = {
                "instance": instance,
                "site": site,
                "protocol": protocol,
            }
            subject = render_to_string("waitinglist/survey_invite_subject.txt", ctx)
            subject = subject.strip()
            message = render_to_string("waitinglist/survey_invite_body.txt", ctx)
            EmailMessage(
                subject,
                message,
                to=[entry.email],
                from_email=settings.WAITINGLIST_SURVEY_INVITE_FROM_EMAIL
            ).send()



import datetime

from django.utils import timezone

from django.contrib.auth.models import User

from account.models import SignupCode

from waitinglist.models import WaitingListEntry


def stats():
    return {
        "waiting_list_entries": WaitingListEntry.objects.count(),
        "waitinglist_added_last_seven_days": WaitingListEntry.objects.filter(created__gt=timezone.now() - datetime.timedelta(days=7)).count(),
        "waitinglist_added_last_thirty_days": WaitingListEntry.objects.filter(created__gt=timezone.now() - datetime.timedelta(days=30)).count(),
        "waiting_list_entries_to_invite": WaitingListEntry.objects.exclude(email__in=SignupCode.objects.values("email")).exclude(email__in=User.objects.values("email")).count()
    }
